import { world, ItemStack, system, Vector } from "@minecraft/server"

const dimensions = ['overworld', 'nether', 'the_end']
let admin = false
let itemMovementTicks = 0
let itemsInMovement = []
let worldEntities = []

function setPermutations(x, y, z, blockName, properties, values) {
    const props = { ...properties };
    for (const [key, val] of Object.entries(values)) {
        if (props.hasOwnProperty(key)) {
            props[key] = val;
        }
    }
    const propStrings = Object.entries(props).map(([key, val]) => `"${key}": ${val}`);
    return `setblock ${x} ${y} ${z} ${blockName} [ ${propStrings.join(", ")} ]`;
}

const worldTicks = () => system.run(() => {
    const t = system.currentTick
    if (t % 4 === 0) {
        for (const player of world.getPlayers()) {
            const item = player.getComponent("inventory").container.getItem(player.selectedSlot)
            const { x, y, z } = player.location
            try {
                const block = world.getDimension(player.dimension.id).getBlock({ x: Math.floor(x), y: Math.floor(y + 1.5), z: Math.floor(z) }).typeId
                if (block === 'effect99:conveyor_tube_up' || block === 'effect99:conveyor_tube_down' || block === 'effect99:conveyor_tube_station' || block === 'effect99:conveyor_tube_station_up' || block === 'effect99:conveyor_tube_up_off' || block === 'effect99:conveyor_tube_down_off') {
                    player.runCommandAsync('playanimation @s animation.conveyor.tube')
                } else if (block === 'effect99:conveyor_tube_north' || block === 'effect99:conveyor_tube_south' || block === 'effect99:conveyor_tube_east' || block === 'effect99:conveyor_tube_west' || block === 'effect99:conveyor_tube_north_off' || block === 'effect99:conveyor_tube_south_off' || block === 'effect99:conveyor_tube_east_off' || block === 'effect99:conveyor_tube_west_off') {
                    player.runCommandAsync('playanimation @s animation.conveyor.tube_sides')
                }
                if (player.isSneaking && world.getDimension(player.dimension.id).getBlock({ x: Math.floor(x), y: Math.floor(y - 1), z: Math.floor(z) }).hasTag('not_full_storage') && player.level >= 1) {
                    const permutation = world.getDimension(player.dimension.id).getBlock({ x: Math.floor(x), y: Math.floor(y - 1), z: Math.floor(z) }).permutation
                    player.runCommandAsync(setPermutations(Math.floor(x), Math.floor(y - 1), Math.floor(z), permutation.type.id, permutation.getAllProperties(), { "xp:level": permutation.getProperty('xp:level') + 1 }))
                    player.runCommandAsync('xp -1l @s')
                }
            } catch (e) { }
            try {
                if (item.typeId === 'effect99:item_hoover') {
                    if (item.getLore()[0]) {
                        if (Number(item.getLore()[0].match(/[\d.]+/g)[0])) {
                            player.onScreenDisplay.setActionBar(item.getLore().reduce((x, y) => (x + '\n     ' + y)))
                        }
                    } else {
                        item.setLore(['§lBattery 0‰', '§l§cOff'])
                        player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                    }
                    const energy = Number(item.getLore()[0].match(/[\d.]+/g)[0])
                    if (item.getLore()[1] === ('§l§aOn')) {
                        let inUse = false
                        let maxPower = false
                        const { x: vx, y: vy, z: vz } = player.getBlockFromViewDirection({ maxDistance: 12, includeLiquidBlocks: true })
                        for (const entity of world.getDimension('overworld').getEntities({ location: { x: vx, y: vy + 1, z: vz }, maxDistance: 8 })) {
                            if (entity.getComponent('minecraft:item')) {
                                inUse = true
                                const { x: ex, y: ey, z: ez } = entity.location
                                const velocity = new Vector(x - ex, y + 1 - ey, z - ez).normalized()
                                if (world.getDimension('overworld').getBlock({ x: ex, y: ey - 1, z: ez }).typeId === 'minecraft:air') {
                                    try { entity.applyImpulse({ x: velocity.x * 0.5, y: velocity.y * 0.5, z: velocity.z * 0.5 }) } catch (e) { }
                                } else {
                                    try { entity.applyImpulse({ x: velocity.x * 0.38, y: velocity.y * 0.38, z: velocity.z * 0.38 }) } catch (e) { }
                                }
                            }
                        }
                        for (let i = -2; i <= 2; i++) {
                            for (let j = -1; j <= 1; j++) {
                                for (let k = -2; k <= 2; k++) {
                                    const block = player.dimension.getBlock({ x: vx + i, y: vy + j, z: vz + k });
                                    try {
                                        const permutation = block.permutation
                                        if (permutation.getProperty('growth') === 7) {
                                            block.dimension.runCommandAsync(`setblock ${vx + i} ${vy + j} ${vz + k} ${block.typeId} 0 destroy`)
                                            inUse = true
                                            break
                                        }
                                    } catch (e) { }
                                    if (block.typeId === 'minecraft:sand' || block.typeId === 'minecraft:gravel') {
                                        block.dimension.runCommandAsync(`setblock ${vx + i} ${vy + j} ${vz + k} air 0 destroy`)
                                        inUse = true
                                        maxPower = true
                                        break
                                    }
                                }
                            }
                        }
                        if (inUse) {
                            if (energy) {
                                item.setLore([`§lBattery ${(energy >= 1 ? maxPower ? energy - 1 : energy - 0.1 : energy - 0.1).toFixed(1)}‰`, '§l§aOn'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                            } else {
                                item.setLore([`§lBattery 0‰`, '§l§cOff'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                            }
                        }
                    } else if (energy === 0) {
                        let battery = 0
                        for (let i = 0; i < 36; i++) {
                            const inventory = player.getComponent('inventory').container
                            const { typeId, amount } = inventory.getSlot(i)
                            switch (typeId) {
                                case 'effect99:small_battery_i': recharge(10); break
                                case 'effect99:small_battery_ii': recharge(20); break
                                case 'effect99:small_battery_iii': recharge(30); break
                                case 'effect99:small_battery_iv': recharge(40); break
                                case 'effect99:small_battery_v': recharge(50); break
                                case 'effect99:small_battery_vi': recharge(60); break
                                case 'effect99:small_battery_vii': recharge(70); break
                                case 'effect99:small_battery_viii': recharge(80); break
                                case 'effect99:small_battery_ix': recharge(90); break
                                case 'effect99:small_battery_x': recharge(100); break
                            }
                            function recharge(charge) {
                                if (amount > 1) {
                                    inventory.setItem(i, new ItemStack(typeId, amount - 1))
                                }
                                let full = 1
                                for (let j = 0; j < 36; j++) {
                                    const { typeId, amount: count } = inventory.getSlot(j)
                                    if (!typeId) {
                                        inventory.setItem(j, new ItemStack('effect99:battery', 1))
                                        full = 0; break
                                    }
                                    else if (amount > 1 && typeId === 'effect99:battery' && count != 64) {
                                        inventory.setItem(j, new ItemStack('effect99:battery', count + 1))
                                        full = 0; break
                                    }
                                    else if (amount <= 1) {
                                        inventory.setItem(i, new ItemStack('effect99:battery'))
                                        full = 0; break
                                    }
                                }
                                if (full) player.dimension.spawnItem(new ItemStack('effect99:battery', 1), { x: x, y: y + 0.5, z: z })
                                item.setLore([`§lBattery ${charge}‰`, '§l§cOff'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                                battery = 1
                            }
                        }
                        if (!battery) player.onScreenDisplay.setActionBar('§l§‰Place a small\nbattery in inventory')
                    }
                }
            } catch (e) { }
            try {
                const itemId = item.typeId
                if (!admin) {
                    if (itemId === 'effect99:powered_cable') give(player, 'effect99:power_cable')
                    else if (itemId === 'effect99:powered_cable_red') give(player, 'effect99:power_cable_red')
                    else if (itemId === 'effect99:powered_cable_blue') give(player, 'effect99:power_cable_blue')
                    else if (itemId === 'effect99:powered_cable_green') give(player, 'effect99:power_cable_green')
                    else if (itemId === 'effect99:powered_reinforced_cable') give(player, 'effect99:reinforced_cable')
                    else if (itemId === 'effect99:powered_reinforced_cable_red') give(player, 'effect99:reinforced_cable_red')
                    else if (itemId === 'effect99:powered_reinforced_cable_blue') give(player, 'effect99:reinforced_cable_blue')
                    else if (itemId === 'effect99:powered_reinforced_cable_green') give(player, 'effect99:reinforced_cable_green')
                    else if (itemId === 'effect99:powered_cable_controller') give(player, 'effect99:power_cable_controller')
                    else if (itemId === 'effect99:quarry_structure' || itemId === 'effect99:small_quarry') give(player, 'effect99:small_quarry_in_box')
                    else if (itemId === 'effect99:medium_quarry_structure' || itemId === 'effect99:medium_quarry') give(player, 'effect99:medium_quarry_in_box')
                    else if (itemId === 'effect99:largue_quarry_structure' || itemId === 'effect99:largue_quarry') give(player, 'effect99:largue_quarry_in_box')
                    else if (itemId === 'effect99:steam_engine') give(player, 'effect99:steam_engine_placer')
                    else if (itemId === 'effect99:steam_generator') give(player, 'effect99:steam_generator_placer')
                    else if (itemId === 'effect99:inefficient_steam_engine') give(player, 'effect99:inefficient_steam_engine_placer')
                    else if (itemId === 'effect99:inefficient_steam_generator' || itemId === 'effect99:isg_blocks') give(player, 'effect99:inefficient_steam_generator_placer')
                    else if (itemId === 'effect99:storage_core') give(player, 'effect99:storage_core_placer')
                    else if (itemId === 'effect99:conveyor_belt_north' || itemId === 'effect99:conveyor_belt_south' || itemId === 'effect99:conveyor_belt_west' || itemId === 'effect99:conveyor_belt_east' || itemId === 'effect99:conveyor_belt_north_off' || itemId === 'effect99:conveyor_belt_south_off' || itemId === 'effect99:conveyor_belt_west_off' || itemId === 'effect99:conveyor_belt_east_off') give(player, 'effect99:conveyor_belt')
                    else if (itemId === 'effect99:conveyor_tube_north' || itemId === 'effect99:conveyor_tube_south' || itemId === 'effect99:conveyor_tube_west' || itemId === 'effect99:conveyor_tube_east' || itemId === 'effect99:conveyor_tube_up' || itemId === 'effect99:conveyor_tube_down' || itemId === 'effect99:conveyor_tube_north_off' || itemId === 'effect99:conveyor_tube_south_off' || itemId === 'effect99:conveyor_tube_west_off' || itemId === 'effect99:conveyor_tube_east_off' || itemId === 'effect99:conveyor_tube_up_off' || itemId === 'effect99:conveyor_tube_down_off') give(player, 'effect99:conveyor_tube')
                    else if (itemId === "effect99:antenna" || itemId === "effect99:antenna_signal" || itemId === "effect99:conveyor_tube_station_up" || itemId === "effect99:power_generating_furnace_up") give(player, 'minecraft:air', true)
                }
            } catch (e) { }
        }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ excludeTypes: ['minecraft:item'] })) {
                    const dimension = world.getDimension(worlds)
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    if (entity.typeId === 'entity:mechanical_arm' || entity.typeId === 'entity:regular_mechanical_arm' || entity.typeId === 'entity:fast_mechanical_arm') {
                        const block = dimension.getBlock({ x: x, y: y, z: z }).permutation
                        switch (entity.getComponent('minecraft:variant').value) {
                            case 0:
                                if (dimension.getBlock({ x: x, y: y, z: z - 1 }).hasTag('need_fuel_1') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x, y, z + 1, dimension, x, y, z - 1)
                                else if (dimension.getBlock({ x: x, y: y, z: z - 2 }).hasTag('need_fuel_2') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x, y, z + 1, dimension, x, y, z - 2)
                                else if (!block.getProperty('cooldown:arm') && !dimension.getBlock({ x: x, y: y, z: z - 1 }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x, y: y, z: z - 2 }).hasTag('adaptable_arm_2')) itemMovement(0, dimension, x, y, z + 1, block, x, y, z, entity)
                                break
                            case 1:
                                if (dimension.getBlock({ x: x, y: y, z: z + 1 }).hasTag('need_fuel_1') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x, y, z - 1, dimension, x, y, z + 1)
                                else if (dimension.getBlock({ x: x, y: y, z: z + 2 }).hasTag('need_fuel_2') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x, y, z - 1, dimension, x, y, z + 2)
                                else if (!block.getProperty('cooldown:arm') && !dimension.getBlock({ x: x, y: y, z: z + 1 }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x, y: y, z: z + 2 }).hasTag('adaptable_arm_2')) itemMovement(1, dimension, x, y, z - 1, block, x, y, z, entity)
                                break
                            case 2:
                                if (dimension.getBlock({ x: x - 1, y: y, z: z }).hasTag('need_fuel_1') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x + 1, y, z, dimension, x - 1, y, z)
                                else if (dimension.getBlock({ x: x - 2, y: y, z: z }).hasTag('need_fuel_2') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x + 1, y, z, dimension, x - 2, y, z)
                                else if (!block.getProperty('cooldown:arm') && !dimension.getBlock({ x: x - 1, y: y, z: z }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x - 2, y: y, z: z }).hasTag('adaptable_arm_2')) itemMovement(2, dimension, x + 1, y, z, block, x, y, z, entity)
                                break
                            case 3:
                                if (dimension.getBlock({ x: x + 1, y: y, z: z }).hasTag('need_fuel_1') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x - 1, y, z, dimension, x + 1, y, z)
                                else if (dimension.getBlock({ x: x + 2, y: y, z: z }).hasTag('need_fuel_2') && !block.getProperty('cooldown:arm'))
                                    fuelProvider(block, x, y, z, entity, x - 1, y, z, dimension, x + 2, y, z)
                                else if (!block.getProperty('cooldown:arm') && !dimension.getBlock({ x: x + 1, y: y, z: z }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x + 2, y: y, z: z }).hasTag('adaptable_arm_2')) itemMovement(3, dimension, x - 1, y, z, block, x, y, z, entity)
                                break
                        }
                    }
                }
            } catch (e) { }
        }
    }
    if (t % 2 === 0) {
        for (let i = 0; i < worldEntities.length; i++) {
            if (worldEntities[i].logCount >= 64) {
                worldEntities[i] = { entityName: worldEntities[i].entityName, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 }
            }
            else if (worldEntities[i].logBlocks[0] || worldEntities[i].otherBlocks[0]) {
                if (worldEntities[i].logBlocks[0]) {
                    function callsSpeed(Num) {
                        for (let ti = 0; ti < Num; ti++) {
                            treeDestroy(i, worldEntities[i].idBlock, worldEntities[i].logBlocks)
                        }
                    }
                    callsSpeed(5)
                }
                if (worldEntities[i].otherBlocks[0]) {
                    function callsSpeed(Num) {
                        for (let ti = 0; ti < Num; ti++) {
                            treeDestroy(i, worldEntities[i].idBlock, worldEntities[i].otherBlocks)
                        }
                    }
                    callsSpeed(78)
                }
            }
            else worldEntities[i] = { entityName: worldEntities[i].entityName, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 }
        }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ excludeTypes: ['minecraft:item'] })) {
                    const dimension = world.getDimension(worlds)
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    if (entity.typeId === 'entity:conveyor_belt_filter') {
                        const permutation = world.getDimension(dimension.id).getBlock({ x: x, y: y, z: z }).permutation
                        for (const itemFilter of entity.getTags()) {
                            for (const filter of dimension.getEntitiesAtBlockLocation({ x: x, y: y, z: z })) {
                                const { x, y, z } = filter.location;
                                const inverted = permutation.getProperty('filter:inverted') ? true : false
                                switch (permutation.getProperty('block:rotation')) {
                                    case 2: filterItem(x, y, z, itemFilter, filter, x, z + 0.8, inverted); break
                                    case 3: filterItem(x, y, z, itemFilter, filter, x, z - 0.8, inverted); break
                                    case 4: filterItem(x, y, z, itemFilter, filter, x + 0.8, z, inverted); break
                                    case 5: filterItem(x, y, z, itemFilter, filter, x - 0.8, z, inverted); break
                                }
                            }
                        }
                        function filterItem(x, y, z, itF, f, ix, iz, inverted) {
                            try {
                                if ((f.getComponent('minecraft:item').itemStack.typeId != itF.trim() &&
                                    inverted || f.getComponent('minecraft:item').itemStack.typeId === itF.trim() &&
                                    !inverted) && y <= Math.floor(y) + 0.2 &&
                                    (x - 0.3 <= Math.floor(x) + 0.5 && x + 0.3 >= Math.floor(x) + 0.5) &&
                                    (z - 0.3 <= Math.floor(z) + 0.5 && z + 0.3 >= Math.floor(z) + 0.5)
                                ) {
                                    f.teleport({ x: ix, y: y + 0.26, z: iz }, dimension, 0, 0)
                                }
                            } catch (e) { }
                        }
                    }
                }
            } catch (e) { }
        }
    }
    for (const worlds of dimensions) {
        try {
            for (const entity of world.getDimension(worlds).getEntities({ type: 'minecraft:item' })) {
                let { x, y, z } = entity.location
                x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                if (y >= -64 && !entity.getTags().some(e => { return e.includes('arm') })) {
                    const block = world.getDimension(worlds).getBlock({ x: x, y: y, z: z })
                    if (block.typeId.includes('conveyor_belt')) {
                        let xDiff = (x + 0.5) - entity.location.x
                        let zDiff = (z + 0.5) - entity.location.z
                        let xVelocity = (block.typeId.includes('west') || block.typeId.includes('east')) ? 0 : xDiff / 16
                        let yVelocity = 0
                        let zVelocity = (block.typeId.includes('north') || block.typeId.includes('south')) ? 0 : zDiff / 16
                        switch (block.typeId) {
                            case 'effect99:conveyor_belt_north': zVelocity += 0.09; break
                            case 'effect99:conveyor_belt_south': zVelocity -= 0.09; break
                            case 'effect99:conveyor_belt_west': xVelocity += 0.09; break
                            case 'effect99:conveyor_belt_east': xVelocity -= 0.09; break
                        }
                        if (block.hasTag('sloping_conveyor_top')) yVelocity = -0.035
                        else if (block.hasTag('sloping_conveyor_down')) yVelocity = 0.13
                        const currentV = entity.getVelocity()
                        const desiredV = { x: xVelocity, y: yVelocity, z: zVelocity }
                        const force = { x: desiredV.x - currentV.x, y: desiredV.y - currentV.y, z: desiredV.z - currentV.z }
                        entity.applyImpulse(force);
                    }
                }
            }
        } catch (e) { }
    }
    if (t % 18 === 0) {
        try {
            for (const player of world.getAllPlayers()) {
                const { x, y, z } = player.location
                let radiationFound = 0
                for (const r of player.dimension.getEntities({ type: 'entity:radiation_explode', location: { x: x, y: y, z: z }, maxDistance: 48 })) {
                    radiationFound = 1
                }
                if (!radiationFound) player.runCommandAsync('fog @s pop radiation')
            }
        } catch (e) { }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ type: 'entity:storage_core' })) {
                    const dimension = world.getDimension(worlds)
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    const core = world.getDimension(dimension.id).getBlock({ x: x, y: y, z: z }).permutation
                    if (entity.typeId === 'entity:storage_core' && core.getProperty('energy:level')) {
                        const positions = [
                            { panel: { x, z: z - 2, itemsEntry: [{ x, z: z - 2 }, { x: x - 1, z: z - 2 }, { x: x + 1, z: z - 2 }] } },
                            { panel: { x, z: z + 2, itemsEntry: [{ x, z: z + 2 }, { x: x - 1, z: z + 2 }, { x: x + 1, z: z + 2 }] } },
                            { panel: { x: x - 2, z, itemsEntry: [{ x: x - 2, z }, { x: x - 2, z: z - 1 }, { x: x - 2, z: z + 1 }] } },
                            { panel: { x: x + 2, z, itemsEntry: [{ x: x + 2, z }, { x: x + 2, z: z - 1 }, { x: x + 2, z: z + 1 }] } }
                        ]
                        let num = 0
                        try {
                            num = entity.getTags()[0].length > 2 ? JSON.parse(entity.getTags()[0]).reduce((prev, { amount }) => prev + amount, 0) - 9999 : 0
                        } catch (e) { }
                        const math = num > 0 && num <= 100000 ? Math.ceil(num / 10000) : 0
                        if (math != core.getProperty('storage:level')) {
                            dimension.runCommandAsync(setPermutations(x, y, z, core.type.id, core.getAllProperties(), { "storage:level": math }))
                        }
                        for (const { panel: { x, z, itemsEntry } } of positions) {
                            const panel = world.getDimension(dimension.id).getBlock({ x: x, y: y - 1, z: z }).permutation
                            let inUse = 0
                            if (panel.getProperty('storage:mode') === 0 && num <= 90000) {
                                for (const { x, z } of itemsEntry) {
                                    for (const item of world.getDimension(dimension.id).getEntitiesAtBlockLocation({ x: x, y: y - 2, z: z })) {
                                        if (item.getComponent('minecraft:item')) {
                                            inUse = 1
                                            const itemStack = item.getComponent('minecraft:item').itemStack
                                            const { typeId, amount, data } = itemStack
                                            let itemSaved = entity.getTags()[0] ? JSON.parse(entity.getTags()[0]) : []
                                            let matched = false
                                            for (const savedItem of itemSaved) {
                                                if (savedItem.typeId === typeId && savedItem.data === data) {
                                                    matched = true
                                                    savedItem.amount += amount
                                                    break
                                                }
                                            }
                                            if (!matched) {
                                                itemSaved.push({ typeId, amount: amount, data })
                                            }
                                            entity.removeTag(`${entity.getTags()[0]}`)
                                            entity.addTag(JSON.stringify(itemSaved))
                                            item.kill()
                                        }
                                    }
                                }
                            }
                            if (panel.getProperty('storage:mode')) {
                                try {
                                    if (entity.getTags()[0].length > 2) {
                                        inUse = 1
                                        let itemSaved = JSON.parse(entity.getTags()[0])
                                        for (const { x, z } of itemsEntry) {
                                            const rNum = Math.floor(Math.random() * itemSaved.length)
                                            const { typeId, amount, data } = itemSaved[rNum]
                                            let maxStack = entity.getComponent('minecraft:mark_variant').value
                                            if (!nonStackableItems.some(e => typeId.includes(e))) { maxStack = Math.min(amount, maxStack) }
                                            else['ender_pearl', 'snowball'].some(e => typeId.includes(e)) ? maxStack = amount >= 16 ? maxStack <= 16 ? maxStack : 16 : Math.min(amount, maxStack) : maxStack = 1
                                            amount === 1 || amount - maxStack === 0 ? itemSaved.splice(rNum, 1) : itemSaved[rNum] = { typeId: typeId, amount: amount - maxStack, data: data }
                                            const item = dimension.spawnItem(new ItemStack(typeId, maxStack), { x: x + 0.5, y: y - 2, z: z + 0.5 })
                                            item.teleport({ x: x + 0.5, y: y - 1.9, z: z + 0.5 }, dimension, 0, 0, false)
                                            switch (panel.getProperty('block:rotation')) {
                                                case 0: item.applyImpulse({ x: 0, y: 0.05, z: 0.15 }); break
                                                case 1: item.applyImpulse({ x: 0, y: 0.05, z: -0.15 }); break
                                                case 2: item.applyImpulse({ x: 0.15, y: 0.05, z: 0 }); break
                                                case 3: item.applyImpulse({ x: -0.15, y: 0.05, z: 0 }); break
                                            }
                                            entity.removeTag(`${entity.getTags()[0]}`)
                                            entity.addTag(JSON.stringify(itemSaved))
                                        }
                                    } else entity.removeTag(`${entity.getTags()[0]}`)
                                } catch (e) { }
                            }
                            try {
                                if (inUse) {
                                    if (entity.getComponent('minecraft:skin_id').value < 100) {
                                        entity.getComponent('minecraft:skin_id').value++
                                    }
                                    else {
                                        let { x, y, z } = entity.location
                                        x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                                        entity.getComponent('minecraft:skin_id').value = 0
                                        const permutation = entity.dimension.getBlock({ x: x, y: y, z: z }).permutation
                                        entity.dimension.runCommandAsync(setPermutations(x, y, z, permutation.type.id, permutation.getAllProperties(), { "energy:level": permutation.getProperty('energy:level') - 1 }))
                                    }
                                }
                            } catch (e) { }
                        }
                    }
                }
            } catch (e) { }
        }
    }
    if (itemMovementTicks) {
        itemMovementTicks -= 1
        itemsInMovement.forEach(({ item, location, dimension, t }, i) => {
            try {
                let { x, y, z } = location
                x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                if (t >= 9) {
                    item.teleport({ x: x + 0.5, y: y + 0.7, z: z + 0.5 }, dimension, 0, 0, false)
                }
                if (item.hasTag('arm0') && t) {
                    item.applyImpulse({ x: 0, y: 0.038, z: -0.024 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm1') && t) {
                    item.applyImpulse({ x: 0, y: 0.038, z: 0.024 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm2') && t) {
                    item.applyImpulse({ x: -0.024, y: 0.038, z: 0 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm3') && t) {
                    item.applyImpulse({ x: 0.024, y: 0.038, z: 0 })
                    itemsInMovement[i].t -= 1
                }
                else {
                    for (const tag of item.getTags()) {
                        item.removeTag(tag)
                    }
                    itemsInMovement.splice(i, 1)
                }
            } catch (e) { }
        })
    }
    worldTicks()
})
worldTicks()

function itemMovement(r, dimension, x, y, z, block, bx, by, bz, entity) {
    for (const item of dimension.getEntitiesAtBlockLocation({ x: x, y: y, z: z })) {
        const { x: ix, y: iy, z: iz } = item.location
        if (item.getComponent('minecraft:item') && iy <= Math.floor(iy) + 0.2 && (ix - 0.3 <= Math.floor(ix) + 0.5 && ix + 0.3 >= Math.floor(ix) + 0.5) && (iz - 0.3 <= Math.floor(iz) + 0.5 && iz + 0.3 >= Math.floor(iz) + 0.5)) {
            switch (r) {
                case 0: item.addTag('arm0'); break
                case 1: item.addTag('arm1'); break
                case 2: item.addTag('arm2'); break
                case 3: item.addTag('arm3'); break
            }
            itemMovementTicks = 20
            dimension.runCommandAsync(setPermutations(bx, by, bz, block.type.id, block.getAllProperties(), { "cooldown:arm": 1 }))
            entity.runCommandAsync(`playanimation @s animation.${entity.typeId.slice(entity.typeId.indexOf(":") + 1)}.move`)
            itemsInMovement.push({ item: item, location: item.location, dimension, t: 10 })
        }
    }
}
function fuelProvider(block, x, y, z, e, ex, ey, ez, dimension, bx, by, bz) {
    let i = 0
    for (const item of e.dimension.getEntitiesAtBlockLocation({ x: ex, y: ey, z: ez })) {
        if (i) break; i++
        const { typeId, amount } = item.getComponent('minecraft:item').itemStack
        if (fuelTypeA.includes(typeId) || fuelTypeB.includes(typeId) || fuelTypeC.includes(typeId) || fuelTypeD.includes(typeId) || (typeId.includes('minecraft:') && keywordFuel.some(i => { return typeId.includes(i) }))) {
            const { x: ix, y: iy, z: iz } = item.location
            const newStack = item.getComponent('minecraft:item').itemStack.clone()
            newStack.amount -= 1
            if (fuelTypeA.includes(typeId)) {
                e.runCommandAsync(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_1>', { x: bx, y: by, z: bz })
                dimension.runCommandAsync(setPermutations(x, y, z, block.type.id, block.getAllProperties(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeB.includes(typeId)) {
                e.runCommandAsync(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_2>', { x: bx, y: by, z: bz })
                dimension.runCommandAsync(setPermutations(x, y, z, block.type.id, block.getAllProperties(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeC.includes(typeId)) {
                e.runCommandAsync(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_3>', { x: bx, y: by, z: bz })
                dimension.runCommandAsync(setPermutations(x, y, z, block.type.id, block.getAllProperties(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeD.includes(typeId) || (typeId.includes('minecraft:') && keywordFuel.some(i => { return typeId.includes(i) }))) {
                e.runCommandAsync(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_4>', { x: bx, y: by, z: bz })
                dimension.runCommandAsync(setPermutations(x, y, z, block.type.id, block.getAllProperties(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
        }
    }
}

world.events.entitySpawn.subscribe(({ entity }) => {
    try {
        if (entity.typeId === 'entity:treecapitator') {
            worldEntities.push({ entityName: entity.id, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 })
            let { x, y, z } = entity.location
            x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
            const blockId = entity.dimension.getBlock({ x: x, y: y + 1, z: z }).typeId
            if ((blockId.includes('log') && !blockId.includes('stripped'))) {
                for (let i = 0; i < worldEntities.length; i++) {
                    if (worldEntities[i].entityName == entity.id) {
                        worldEntities[i].logBlocks.push(`${x} ${y + 1} ${z}`)
                        worldEntities[i].idBlock = blockId
                    }
                }
            }
        }
    } catch (e) { }
})

function treecapitator(block, x, y, z, e) {
    try {
        const blockSides = [[x, y - 1, z], [x + 1, y, z], [x, y, z - 1], [x - 1, y, z], [x, y, z + 1], [x, y + 1, z], [x + 1, y + 1, z + 1], [x + 1, y + 1, z - 1], [x - 1, y + 1, z - 1], [x - 1, y + 1, z + 1]]
        for (const coords of blockSides) {
            const idBlocks = world.getDimension("overworld").getBlock({ x: coords[0], y: coords[1], z: coords[2] }).typeId
            if (idBlocks == block && worldEntities[e].logCount < 64) {
                if (worldEntities[e].logBlocks.every(l => l != `${coords[0]} ${coords[1]} ${coords[2]}`)) {
                    worldEntities[e].logBlocks.push(`${coords[0]} ${coords[1]} ${coords[2]}`)
                    worldEntities[e].logCount += 1
                }
            }
            if ((idBlocks.includes('leaves') || idBlocks.includes('vine')) && (worldEntities[e].otherCount < worldEntities[e].logCount * 6 && worldEntities[e].otherCount < 290)) {
                if (worldEntities[e].otherBlocks.every(o => o != `${coords[0]} ${coords[1]} ${coords[2]}`)) {
                    worldEntities[e].otherBlocks.push(`${coords[0]} ${coords[1]} ${coords[2]}`)
                    worldEntities[e].otherCount += 1
                }
            }
        }
    } catch (e) { }
}
function treeDestroy(e, b, list) {
    try {
        if (worldEntities[e].logBlocks[0] || worldEntities[e].otherBlocks[0]) {
            const x = Number(list[0].slice(0, list[0].indexOf(' ')))
            const y = Number(list[0].slice(list[0].indexOf(' '), list[0].indexOf(' ', list[0].indexOf(' ') + 1)))
            const z = Number(list[0].slice(list[0].indexOf(' ', list[0].indexOf(' ') + 1)))
            //world.getDimension("overworld").runCommandAsync(`title @a actionbar ${b} ${worldEntities[e].logCount} : ${worldEntities[e].otherCount} | ${x} ${y} ${z} | \nIndex Entities: ${e}`)
            treecapitator(b, x, y, z, e)
        }
        world.getDimension("overworld").runCommandAsync(`setblock ${list.splice(0, 1)} air 0 destroy`)
    } catch (e) { }
}

function give(player, item, message = false) {
    player.runCommandAsync(`replaceitem entity @s slot.weapon.mainhand 0 ${item}`)
    if (message) player.onScreenDisplay.setActionBar(`You do not have permission\n to have this item`)
}

world.events.beforeChat.subscribe(chat => {
    if (chat.message === '!admin') {
        chat.cancel = true
        admin = true
        world.sendMessage(`Admin ${chat.sender.name}`)
    } else if (chat.message === '!admin!') {
        chat.cancel = true
        admin = false
        world.sendMessage(`User ${chat.sender.name}`)
    } else if (chat.message === '!addonV') {
        chat.cancel = true
        world.sendMessage(`Add-On Version Beta 0.3.3`)
    }
})

const nonStackableItems = [
    'sword',
    'axe',
    'pickaxe',
    'shovel',
    'hoe',
    'helmet',
    'chestplate',
    'leggings',
    'boots',
    'bucket',
    'bed',
    'ender_pearl',
    'bow',
    'shield',
    'elytra',
    'totem',
    'trident',
    'potion',
    'on_a_stick',
    'fishing_rod',
    'flint_and_steel',
    'shears',
    'snowball',
    'saddle',
    'horse_armor',
    'boat'
]

const fuelTypeA = [
    'minecraft:coal_block',
    'minecraft:blaze_rod',
    'minecraft:lava_bucket'
]

const fuelTypeB = [
    'minecraft:coal',
    'minecraft:charcoal',
    'minecraft:dried_kelp_block',
    'minecraft:blaze_powder'
]

const fuelTypeC = [
    'minecraft:planks',
    'minecraft:log',
    'minecraft:log2',
    'minecraft:planks',
    'minecraft:wood',
    'minecraft:stripped_acacia_log',
    'minecraft:stripped_birch_log',
    'minecraft:stripped_dark_oak_log',
    'minecraft:stripped_jungle_log',
    'minecraft:stripped_oak_log',
    'minecraft:stripped_spruce_log',
    'minecraft:mangrove_wood',
    'minecraft:wooden_door',
    'minecraft:wool',
    'minecraft:lectern',
    'minecraft:chest',
    'minecraft:trapped_chest',
    'minecraft:barrel',
    'minecraft:composter',
    'minecraft:crafting_table',
    'minecraft:bookshelf',
    'minecraft:noteblock'
]

const fuelTypeD = [
    'minecraft:stick',
    'minecraft:ladder',
    'minecraft:deadbush',
    'minecraft:sapling'
]

const keywordFuel = [
    'jungle',
    'acacia',
    'birch',
    'dark_oak',
    'spruce',
    'mangrove',
    'oak',
    'wooden',
    'leaves',
    'azalea'
]